#include <vector>
#include <cstdio>

using std::printf;
using std::putchar;
using std::fopen;
using std::fread;
using std::fclose;

static class ROM: public std::vector<unsigned char>
{
public:
    ROM()
    {
        resize(0x20000);
        FILE*fp = fopen("megaman1.nes", "rb");
        fseek(fp, 0x10, SEEK_SET);
        fread(&operator[](0), 1, size(), fp);
        fclose(fp);
    }
} ROM;

struct Char
{
    unsigned char c;
public:
    Char() :c(0) { }
    Char(char ch) { c=ch; }
    Char& operator=(char ch) { c=ch; return *this; }
    
    void Dump() const
    {
        /* Classify types.
         *
         *
         *
         */
        switch(c)
        {
            // air:
            case 0x00:
            case 0x01:
            case 0x23:
            case 0x24:
            case 0x25:
            case 0x26:
            case 0x2B:
            case 0x2C:
            case 0x34:
            case 0x35:
            case 0x3C:
            default:
                putchar('.');
                break;
            case 0x51:
            case 0x52:
            case 0x53:
            case 0x62:
            case 0x68:
            case 0x69:
                putchar('#');
                break;
            case 0x38:
            case 0x39:
            case 0x3A:
            case 0x3B:
                putchar('G');
                break;
            case 0x82:
                putchar('|');
                break;
            case 0xC4:
                putchar('^');
        }
        //printf("%02X",c);
    }
};

struct Block
{
    Char b[4];
    unsigned blockid;
public:
    void Dump()  const
    {
        printf("%3u", blockid);
    }
};

static class Blocks: public std::vector<Block>
{
public:
    Blocks()
    {
        Blocks& t = *this;
        const unsigned base = 0x0000;
        resize(110);
        for(unsigned a=0; a<size(); ++a)
        {
            t[a].b[0] = ROM[base + a*4+0];
            t[a].b[1] = ROM[base + a*4+1];
            t[a].b[2] = ROM[base + a*4+2];
            t[a].b[3] = ROM[base + a*4+3];
            t[a].blockid = a;

            unsigned char palettebyte = ROM[0x300 + a];

            for(unsigned b=0; b<4; ++b)
            {
                const unsigned trans[4] = {0,2,1,3};
                unsigned transt = trans[b];
                
                char bit1 = 1&(palettebyte >> ((transt*2) +0));
                char bit2 = 1&(palettebyte >> ((transt*2) +1));
/*
                printf("%02X has palette %u\n",
                    t[a].b[b].c,
                    
                    bit1*2 + bit2);
*/
            }
        }
    }
} Blocks;

struct Room
{
    Block blocks[8*8];
public:
    Room()
    {
    }
    void Dump() const
    {
        for(unsigned y=0; y<16; ++y)
        {
            printf(" \"");
            for(unsigned x=0; x<16; ++x)
            {
                blocks[(x/2)*8+(y/2)].
                     b[(x%2)*2+(y%2)].Dump();
            }
            printf("\"\n");
        }
    }
};

static class Rooms: public std::vector<Room>
{
public:
    Rooms()
    {
        Rooms& t = *this;
    
        const unsigned base = 0x0C30; //  location of the pointesr
        resize(32);
        for(unsigned a=0; a<size(); ++a)
        {
            unsigned pointer = ROM[base + a*2] + ROM[base + a*2+1]*256;
            pointer -= 0x8000;
            
            for(unsigned n=0; n<64; ++n)
                t[a].blocks[n] = Blocks[ROM[pointer + n]];
        }
    }
} Rooms;

struct Screen
{
    Room room;
    unsigned roomid;
    enum { right,up,down, end } dir;
    bool shutter;
public:
    Screen(): dir(end), shutter(false),roomid(999)
    {
    }
    void Dump() const
    {
        char flag = '?';
        switch(dir)
        {
            case right: flag='R'; break;
            case    up: flag='U'; break;
            case  down: flag='D'; break;
            case end: flag='e'; break;
        }
        if(shutter) flag='S';

        printf("{\n");
        room.Dump();
        printf(",'%c'},\n", flag);
    }
};

static class Level: public std::vector<Screen>
{
public:
    Level()
    {
        Level& t = *this;
    
        unsigned screen = 0;
        unsigned pos_flags = 0xC71;
        unsigned pos_rooms = 0xC00;
        for(;;)
        {
            unsigned char b = ROM[pos_flags++];
            
            unsigned n_screens = (b & 0x1F);
            for(unsigned a=0; a<=n_screens; ++a, ++screen)
            {
                resize(screen+1);
            redo:
                unsigned char c = ROM[pos_rooms++];
                if(c >= Rooms.size())
                {
                    printf("c=%02X, ignoring\n", c);
                    goto redo;
                }                
                t[screen].room   = Rooms[c];
                t[screen].roomid = c;
                t[screen].dir    = Screen::right;
                
                if(a == n_screens)
                {
                    switch(b & 0xE0)
                    {
                        case 0x00: t[screen].dir = Screen::end; break;
                        case 0x20: t[screen].shutter = true; break;
                        case 0x40: t[screen].dir = Screen::down; break;
                        case 0x80: t[screen].dir = Screen::up; break;
                        default:
                            printf("Dir %02X?\n", b & 0xE0);
                    }
                }
            }
            
            if(!b) break;
        }
    }
    void Dump() const
    {
        const Level& t = *this;
        for(unsigned a=0; a<size(); ++a)
        {
            t[a].Dump();
        }
    }
} Level;


int main(void)
{
    Level.Dump();
    return 0;
}
