#include <vector>
#include <cstdio>

using std::printf;
using std::putchar;
using std::fopen;
using std::fread;
using std::fclose;

static const struct
{
    const char* ascii;
    char type;
} leveldata[] =
{
{
 "................"
 "................"
 "................"
 "................"
 "........GG......"
 "........GG......"
 "..............##"
 "................"
 "..........GG...."
 "..........GG...."
 "........GGGG...."
 "........GGGG...."
 "................"
 "................"
 "################"
 "################"
,'R'},
{
 "..........######"
 "..........######"
 "................"
 "................"
 "........GG......"
 "........GG......"
 "........##......"
 "........##......"
 "................"
 "................"
 ".......#########"
 "........########"
 "##.............."
 "####............"
 "################"
 "################"
,'R'},
{
 "................"
 "................"
 "................"
 "................"
 "....##.........."
 "....##.........."
 "..######GG..####"
 "..######GG......"
 "|#########......"
 "|...########|..."
 "|...########|..."
 "###|##......|..."
 "...|........|..."
 "...|....########"
 "####........####"
 "####^^^^^^..####"
,'R'},
{
 "............#|##"
 "............#|##"
 ".............|##"
 "..............##"
 "..............##"
 "...........#####"
 "#...#|##....####"
 "#...#|##....####"
 "#...#|##....####"
 "#...#|#####|####"
 "#.....##...|####"
 "#.....##...|####"
 "########...|####"
 "################"
 "################"
 "################"
,'U'},
{
 ".............|##"
 ".............|##"
 ".............|##"
 "..........######"
 "..........######"
 "........########"
 "........########"
 "......##########"
 "..|#..##########"
 "..|#..........##"
 "..|...........##"
 "..|...........##"
 "..|...........##"
 "##########..#|##"
 "##########..#|##"
 "##########..#|##"
,'U'},
{
 ".............|##"
 ".............|##"
 ".............|##"
 "............####"
 "............####"
 "..|#......######"
 "..|#......######"
 "..|.....#|######"
 "..|.....#|######"
 "..|......|....##"
 "..|......|....##"
 "..##.....|....##"
 "..##.....|....##"
 "..###...#####|##"
 "..###...#####|##"
 "..###...#####|##"
,'U'},
{
 "...|..##########"
 "...|...#########"
 "...|......######"
 "...|.......#####"
 "...|........####"
 "...#####....####"
 "............####"
 "............####"
 "...#|###########"
 "....|.......####"
 "....|.........##"
 "....|.........##"
 "....|.........##"
 "...####..####|##"
 "...####..####|##"
 "...####..####|##"
,'U'},
{
 "................"
 "................"
 "................"
 "................"
 "................"
 "................"
 "................"
 "................"
 "................"
 "................"
 "................"
 "................"
 "............####"
 "...|############"
 "...|############"
 "...|############"
,'R'},
{
 "................"
 "................"
 "................"
 "................"
 "................"
 "................"
 "........GG......"
 ".##.....GG......"
 "......####......"
 "......####..##.."
 "....######..##.."
 "....######..####"
 "##########..####"
 "##########..####"
 "##########..####"
 "##########..####"
,'R'},
{
 "............#|##"
 "............#|##"
 ".............|##"
 ".............|##"
 "....#|#..#######"
 "....#|#..#######"
 "....#|......####"
 "....#|......####"
 "....#|......####"
 "....#|###...####"
 "...........#####"
 "...........#####"
 "################"
 "################"
 "################"
 "################"
,'U'},
{
 "##...........|##"
 "##...........|##"
 "##...........|##"
 "##...........|##"
 "####........####"
 "##............##"
 "##......##....##"
 "####|#......####"
 "##..|.........##"
 "##..|.........##"
 "##..|.GG......##"
 "##..|.GG......##"
 "#############|##"
 "#############|##"
 "#############|##"
 "#############|##"
,'U'},
{
 "##|...........##"
 "##|...........##"
 "##|...........##"
 "##|...........##"
 "##|...........##"
 "##|...........##"
 "##|.........####"
 "##|.........####"
 "##GG......######"
 "##GG......######"
 "######..########"
 "######........##"
 "######........##"
 "#############|##"
 "#############|##"
 "#############|##"
,'U'},
{
 "###|############"
 "###|############"
 "##.|..........##"
 "##.|..........##"
 "##............##"
 "##............##"
 "####|.......####"
 "##..|.......####"
 "##..|.GG....####"
 "##..|.GG....####"
 "##..############"
 "##..############"
 "##..############"
 "##|#############"
 "##|#############"
 "##|#############"
,'U'},
{
 "................"
 "................"
 "................"
 "................"
 "##.............."
 "##.............."
 "##.............."
 "##.............."
 "##.............."
 "##.............."
 "##.............."
 "##.............."
 "##............##"
 "###|############"
 "###|############"
 "###|############"
,'R'},
{
 "................"
 "................"
 "................"
 "................"
 "................"
 "................"
 "......##########"
 "......##########"
 "..##############"
 "..##############"
 "################"
 "################"
 "################"
 "################"
 "################"
 "################"
,'R'},
{
 "................"
 "................"
 "................"
 "................"
 "................"
 "................"
 "................"
 "................"
 "############...."
 "############...."
 "##.............."
 "##.............."
 "############|#.."
 "############|#.."
 "############|#.."
 "############|#.."
,'D'},
{
 "######......|..."
 "#####.......|..."
 "#####.......|..."
 "######|#######.."
 "######|#######.."
 "####............"
 "####............"
 "##.............."
 "##.............."
 "##....##..##...."
 "##.............."
 "##.............."
 "##.............."
 "######|#........"
 "######|#........"
 "######|#........"
,'D'},
{
 "######|#........"
 "######|#........"
 "######.........."
 "######.......^^."
 "########.....##."
 "########........"
 "####............"
 "####............"
 "##.............."
 "##.....##....##."
 "##.............."
 "##.............."
 "##.............."
 "##|#####^^^^^^^^"
 "##|#############"
 "##|#############"
,'D'},
{
 "##|............."
 "##|............."
 "##|............."
 "##|............."
 "##.............."
 "##.............."
 "##.............."
 "##.............."
 "##GGGG.........."
 "##GGGG.........."
 "####GGGG........"
 "####GGGG........"
 "################"
 "################"
 "################"
 "################"
,'R'},
{
 ".............###"
 ".............###"
 ".............###"
 ".............###"
 ".............###"
 ".............###"
 "................"
 "................"
 "................"
 "................"
 "..........######"
 "..........######"
 "################"
 "################"
 "################"
 "################"
,'S'},
{
 "................"
 "................"
 "................"
 "................"
 "################"
 "################"
 "................"
 "................"
 "................"
 "................"
 "################"
 "################"
 "................"
 "................"
 "................"
 "................"
,'R'},
{
 "................"
 "................"
 "................"
 "................"
 "################"
 "################"
 "................"
 "................"
 "................"
 "................"
 "################"
 "################"
 "................"
 "................"
 "................"
 "................"
,'R'},
{
 "................"
 "................"
 "................"
 "................"
 "################"
 "################"
 "................"
 "................"
 "................"
 "................"
 "################"
 "################"
 "................"
 "................"
 "................"
 "................"
,'S'},
{
 "################"
 "################"
 "#..............#"
 "#..............#"
 "#..............#"
 "#..............#"
 "...............#"
 "...............#"
 "...............#"
 "...............#"
 "##GGGG.........#"
 "##GGGG.........#"
 "################"
 "################"
 "################"
 "################"
,'e'},
};
static const unsigned n_screens = sizeof(leveldata) / sizeof(*leveldata);


static class ROM: public std::vector<unsigned char>
{
public:
    ROM()
    {
        resize(0x20000);
        FILE*fp = fopen("megaman1.nes", "rb");
        fseek(fp, 0x10, SEEK_SET);
        fread(&operator[](0), 1, size(), fp);
        fclose(fp);
    }
    ~ROM()
    {
        FILE*fp = fopen("tmp.nes", "wb");
        fprintf(fp, "NES%c%c%c!DiskDude!", 26,8,0);
        fwrite(&operator[](0), 1, size(), fp);
        fclose(fp);
    }
} ROM;

struct Char
{
    unsigned char tile;    //0-FF
    unsigned char palette; //0-3
public:
    bool operator== (const Char& b) const
    {
        return tile==b.tile && palette==b.palette;
    }
};

struct Block
{
    Char c[4];
public:
    void Write(unsigned offs, unsigned color_offs) const
    {
        for(unsigned a=0; a<4; ++a)
            ROM[offs + a] = c[a].tile;
            
        unsigned palette =
            (c[0].palette<<0) |
            (c[2].palette<<2) |
            (c[1].palette<<4) |
            (c[3].palette<<6);
            
        ROM[color_offs] = palette;
    }
    
    bool operator==(const Block& b) const
    {
        return c[0] == b.c[0]
            && c[1] == b.c[1]
            && c[2] == b.c[2]
            && c[3] == b.c[3];
    }
};
static class Blocks: public std::vector<Block>
{
public:
    Blocks()
    {
        /*
        for(unsigned a=0; a<110; ++a)
        {
            Block tmp;
            tmp.c[0] = ROM[0x0000 + a*4+0];
            tmp.c[1] = ROM[0x0000 + a*4+1];
            tmp.c[2] = ROM[0x0000 + a*4+2];
            tmp.c[3] = ROM[0x0000 + a*4+3];
            push_back(tmp);
        }
        */
        
        /*
        Block zeroblock;
        zeroblock.c[0] = 0;
        zeroblock.c[1] = 1;
        zeroblock.c[2] = 2;
        zeroblock.c[3] = 3;
        push_back(zeroblock);
        */
    }
    unsigned Get(const Block& tmp)
    {
        // Return the index of the block.
        // If already exists, return existing index.
        Blocks& t = *this;
        for(unsigned a=0; a<size(); ++a)
            if(t[a] == tmp)
                return a;
        /*
        for(unsigned a=0; a<size(); ++a)
        {
            printf("[%3u]", a);
            printf("%02X %02X %02X %02X\n",
                t[a].c[0],t[a].c[1],t[a].c[2],t[a].c[3]);
        }
        */
        /*
        printf("Adding %02X %02X %02X %02X\n",
            tmp.c[0],tmp.c[1],tmp.c[2],tmp.c[3]);
        */
        tmp.Write(0x0000 + size()*4,
                  0x0300 + size()*1);
        push_back(tmp);
        return size()-1;
    }
} Blocks;

struct Room
{
    unsigned blocks[8*8];
public:
    void Write(unsigned ptroffs, unsigned offs) const
    {
        ROM[ptroffs+0] = ((offs+0x8000) & 255);
        ROM[ptroffs+1] = ((offs+0x8000) >> 8);
        for(unsigned a=0; a<64; ++a) ROM[offs+a] = blocks[a];
    }
    
    void CreateFrom(const char* desc)
    {
        char inmap[16*16];
        Char outmap[16*16];
        for(unsigned a=0; desc[a]; ++a) inmap[a] = desc[a];
        
        // Convert to tiles
        for(unsigned y=0; y<16; ++y)
            for(unsigned x=0; x<16; ++x)
            {
                unsigned char c   = 0;
                unsigned char pal = 0;
                
                /* Cutman palettes:
                 *   
                 * 0 = solid
                 * 1 = blacks
                 * 2 = blues
                 * 3 = blue,gray
                 */
                
                switch(inmap[y*16+x])
                {
                    case '.':
                        c   = 0x01; // background
                        pal = 0x02;
                        break;
                    case '#':
                        if(y > 0 &&
                            (inmap[(y-1)*16+x] != '#')
                          )
                            c = 0x62; // step
                        else
                            c = 0x68 + ((x+y) & 1);
                        pal = 0x00;
                        break;
                    case '|':
                        c   = 0x82; // ladder
                        pal = 0x00;
                        break;
                    case '^':
                        c   = 0xC4; // spike
                        pal = 0x02;
                        break;
                    case 'G':
                        c = 0x38 + (x&1) + (y&1)*2; // gutsblock
                        pal = 0x00;
                        break;
                    default:
                        printf("? '%c'\n", inmap[y*16+x]);
                }
                
                outmap[y*16+x].tile    = c;
                outmap[y*16+x].palette = pal;
            }

        // Convert to blocks
        for(unsigned y=0; y<8; ++y)
            for(unsigned x=0; x<8; ++x)
            {
                Block tmp;
                tmp.c[0] = outmap[(y*2+0)*16 + (x*2+0)];
                tmp.c[1] = outmap[(y*2+1)*16 + (x*2+0)];
                tmp.c[2] = outmap[(y*2+0)*16 + (x*2+1)];
                tmp.c[3] = outmap[(y*2+1)*16 + (x*2+1)];
                blocks[x*8+y] = Blocks.Get(tmp);
            }
    }
    
    bool operator==(const Room& b) const
    {
        for(unsigned a=0; a<64; ++a)
            if(blocks[a] != b.blocks[a])
                return false;
        return true;
    }
};
static class Rooms: public std::vector<Room>
{
public:
    unsigned Get(const Room& tmp)
    {
        // Return the index of the room.
        // If already exists, return existing index.
        Rooms& t = *this;
        for(unsigned a=0; a<size(); ++a)
            if(t[a] == tmp)
            {
                printf("Repeat room %u\n", a);
                return a;
            }
        
        printf("Writing room %u\n", size());
        
        tmp.Write(0x0C30 + size()*2,
                  0x03C0 + size()*64);
        
        push_back(tmp);
        return size()-1;
    }
} Rooms;
static class Level
{
public:
    Level()
    {
        unsigned pos_rooms = 0xC00;
        unsigned pos_flags = 0xC71;
        
        unsigned byte = 0x00;
    
        for(unsigned a=0; a<n_screens; ++a)
        {
            Room room;
            room.CreateFrom(leveldata[a].ascii);
            unsigned room_id = Rooms.Get(room);
            
            ROM[pos_rooms++] = room_id;
            
            switch(leveldata[a].type)
            {
                case 'R': byte += 1; break;
                case 'U':
                    ROM[pos_flags++] = byte | 0x80;
                    byte = 0;
                    break;
                case 'D':
                    ROM[pos_flags++] = byte | 0x40;
                    byte = 0;
                    break;
                case 'S':
                    ROM[pos_flags++] = byte | 0x20;
                    byte = 0;
                    break;
                case 'e':
                    ROM[pos_flags++] = 0x00;
            }
        }
    }
} Level;

int main(void)
{

// 3A = wily3 boss
// 3B = activate enemy
// 3C = small ball
// 3D = small weapon
// 3E = small energy
// 3F = big weapon power
// 44 = end level
// 45 = ?
// 49 = random powerup

    for(unsigned a=0; a<300; ++a)
    {
        unsigned char obj[4] =
        {
            // screen, x, y, type.
            a+2,
            8*16 + 8,
            12*16 + 8,
            0x41
        };
        
        if(a > 2 || a == 0) obj[0] = 0xFF;
        
        for(unsigned b=0; b<4; ++b)
            ROM[0x1A468 + a*4 + b] = obj[b];
    }
    
    unsigned char Palette[16] =
    {
        0x0F, 0x29,0x27,0x06,
        0x0F, 0x30,0x2B,0x06,
        0x0F, 0x30,0x2B,0x01,
        0x0F, 0x30,0x2B,0x06
    };
    
    for(unsigned a=0; a<16; ++a)
    {
        ROM[0x0CA0 + a] = Palette[a];
    }
}
